@@ -26,6 +26,7 @@
 #endif
 
 #include "exchange-mapi-cal-recur-utils.h"
+#include <libecal/e-cal-util.h>
 
 /* Reader/Writer versions */
 #define READER_VERSION	0x3004
@@ -76,6 +77,18 @@ struct ExceptionInfo {
 };
 #endif
 
+/* Override flags defining what fields might be found in ExceptionInfo */
+#define ARO_SUBJECT 0x0001
+#define ARO_MEETINGTYPE 0x0002
+#define ARO_REMINDERDELTA 0x0004
+#define ARO_REMINDER 0x0008
+#define ARO_LOCATION 0x0010
+#define ARO_BUSYSTATUS 0x0020
+#define ARO_ATTACHMENT 0x0040
+#define ARO_SUBTYPE 0x0080
+#define ARO_APPTCOLOR 0x0100
+#define ARO_EXCEPTIONAL_BODY 0x0200
+
 static icalrecurrencetype_weekday
 get_ical_weekstart (uint32_t fdow)
 {
@@ -234,11 +247,11 @@ check_calendar_type (guint16 type)
 }
 
 gboolean
-exchange_mapi_cal_util_bin_to_rrule (GByteArray *ba, ECalComponent *comp)
+exchange_mapi_cal_util_bin_to_rrule (GByteArray *ba, ECalComponent *comp, GSList **extra_detached)
 {
 	struct icalrecurrencetype rt;
 	guint16 flag16;
-	guint32 flag32;
+	guint32 flag32, writer_version;
 	guint8 *ptr = ba->data;
 	gint i;
 	GSList *exdate_list = NULL;
@@ -632,7 +645,9 @@ exchange_mapi_cal_util_bin_to_rrule (GByteArray *ba, ECalComponent *comp)
 	/* number of changed exceptions */
 	flag32 = *((guint32 *)ptr);
 	ptr += sizeof (guint32);
-	/* FIXME: Parse modified instances */
+	/* For each changed exception, there will be a corresponding
+          ExceptionInfo below.  So at present we don't need to do
+          anything with the information here beyond skipping it */
 	if (flag32)
 		ptr += flag32 * sizeof (guint32);
 
@@ -654,10 +669,12 @@ exchange_mapi_cal_util_bin_to_rrule (GByteArray *ba, ECalComponent *comp)
 		return FALSE;
 
 	/* some constant */
-	flag32 = *((guint32 *)ptr);
+	writer_version = *((guint32 *)ptr);
 	ptr += sizeof (guint32);
-	if (flag32 != WRITER_VERSION2)
-		return FALSE;
+	/* It should be set, but not must. It can be, technically, any value.
+	   Seen were 0x3006, 0x3008, 0x3009. It affects format of extended exception info
+	if (writer_version != WRITER_VERSION2)
+		return FALSE; */
 
 	/* start time in mins */
 	flag32 = *((guint32 *)ptr);
@@ -667,24 +684,6 @@ exchange_mapi_cal_util_bin_to_rrule (GByteArray *ba, ECalComponent *comp)
 	flag32 = *((guint32 *)ptr);
 	ptr += sizeof (guint32);
 
-	/* modified exceptions */
-	flag16 = *((guint16 *)ptr);
-	ptr += sizeof (guint16);
-	if (flag16 != 0x0)
-		return FALSE;
-
-	/* reserved block1 size - has to be 0 */
-	flag32 = *((guint32 *)ptr);
-	ptr += sizeof (guint32);
-	if (flag32 != 0x0)
-		return FALSE;
-
-	/* reserved block2 size - has to be 0 */
-	flag32 = *((guint32 *)ptr);
-	ptr += sizeof (guint32);
-	if (flag32 != 0x0)
-		return FALSE;
-
 	/* Set the recurrence */
 	{
 		GSList l;
@@ -698,11 +697,222 @@ exchange_mapi_cal_util_bin_to_rrule (GByteArray *ba, ECalComponent *comp)
 	/* FIXME: this also has modified instances */
 	e_cal_component_set_exdate_list (comp, exdate_list);
 
-	g_print ("\n== MAPI to ICAL == The recurrence blob data is as follows:\n");
-	for (i = 0; i < ba->len; ++i)
-		g_print ("0x%02X ", ba->data[i]);
-	g_print("\n== End of stream ==\n");
+	/* modified exceptions, an ExceptionCount sized list of
+	   ExceptionInfo instances */
+	flag16 = *((guint16 *)ptr);
+	ptr += sizeof (guint16);
+	if (flag16 && extra_detached) {
+		gint count = flag16;
+		guint16 *overrideflags = g_new0 (guint16, count);
+		ECalComponent **detached = g_new0 (ECalComponent *, count);
+		uint32_t starttime, endtime, origtime;
+		struct icaltimetype tt;
+		ECalComponentDateTime edt;
+		ECalComponentRange rid;
 
+		e_cal_component_commit_sequence (comp);
+
+		for (i = 0; i < count; i++) {
+			/* ExceptionInfo.StartTime */
+			starttime = *((guint32 *)ptr);
+			ptr += sizeof (guint32);
+
+			/* ExceptionInfo.EndTime */
+			endtime = *((guint32 *)ptr);
+			ptr += sizeof (guint32);
+
+			/* ExceptionInfo.OriginalStartDate */
+			origtime = *((guint32 *)ptr);
+			ptr += sizeof (guint32);
+
+			/* make a shallow clone of comp */
+			detached[i] = e_cal_component_clone (comp);
+
+			tt = icaltime_from_timet_with_zone (convert_recurrence_minutes_to_timet (origtime), 0, 0);
+			rid.type = E_CAL_COMPONENT_RANGE_SINGLE;
+			rid.datetime.value = &tt;
+			rid.datetime.tzid = "UTC";
+			e_cal_component_set_recurid (detached[i], &rid);
+
+			tt = icaltime_from_timet_with_zone (convert_recurrence_minutes_to_timet (starttime), 0, 0);
+			edt.value = &tt;
+			edt.tzid = "UTC";
+			e_cal_component_set_dtstart (detached[i], &edt);
+
+			tt = icaltime_from_timet_with_zone (convert_recurrence_minutes_to_timet (endtime), 0, 0);
+			edt.value = &tt;
+			edt.tzid = "UTC";
+			e_cal_component_set_dtend (detached[i], &edt);
+
+			e_cal_component_set_rdate_list (detached[i], NULL);
+			e_cal_component_set_rrule_list (detached[i], NULL);
+			e_cal_component_set_exdate_list (detached[i], NULL);
+			e_cal_component_set_exrule_list (detached[i], NULL);
+
+			/* continue parsing stuff we don't need, because we need to
+			   get to the next ExceptionInfo object or back out to the
+			   containing AppointmentRecurrencePattern object */
+
+			/* ExceptionInfo.OverrideFlags */
+			overrideflags[i] = *((guint16 *) ptr);
+			ptr += sizeof (guint16);
+
+			if (overrideflags[i] & ARO_SUBJECT) {
+				ECalComponentText text = { 0 };
+				gchar *str;
+
+				/* ExceptionInfo.SubjectLength, ExceptionInfo.SubjectLength2
+				   and ExceptionInfo.Subject */
+				ptr += sizeof (guint16);
+				flag16 = *(guint16 *)ptr; /* use SubjectLength2 */
+				ptr += sizeof (guint16);
+				/* note a discrepency in MS-OXOCAL here, which suggests that
+				   Subject is actually 2 bytes */
+
+				str = g_strndup ((const gchar *) ptr, flag16);
+				text.value = str;
+				e_cal_component_set_summary (detached[i], &text);
+				g_free (str);
+
+				ptr += flag16;
+			}
+
+			if (overrideflags[i] & ARO_MEETINGTYPE) {
+				/* ExceptionInfo.MeetingType */
+				ptr += sizeof (guint32);
+			}
+
+			if (overrideflags[i] & ARO_REMINDERDELTA) {
+				/* ExceptionInfo.ReminderDelta */
+				ptr += sizeof (guint32);
+			}
+
+			if (overrideflags[i] & ARO_REMINDER) {
+				/* ExceptionInfo.ReminderSet */
+				ptr += sizeof (guint32);
+			}
+
+			if (overrideflags[i] & ARO_LOCATION) {
+				gchar *str;
+
+				/* ExceptionInfo.LocationLength, ExceptionInfo.LocationLength2
+				   and ExceptionInfo.Location */
+				ptr += sizeof (guint16);
+				flag16 = *(guint16 *) ptr; /* use LocationLength2 */
+				ptr += sizeof (guint16);
+				/* note a discrepency in MS-OXOCAL here, which suggests that
+				   Location is actually 4 bytes */
+
+				str = g_strndup ((const gchar *) ptr, flag16);
+				e_cal_component_set_location (detached[i], str);
+				g_free (str);
+
+				ptr += flag16;
+			}
+
+			if (overrideflags[i] & ARO_BUSYSTATUS) {
+				/* ExceptionInfo.BusyStatus */
+				ptr += sizeof (guint32);
+			}
+
+			if (overrideflags[i] & ARO_ATTACHMENT) {
+				/* ExceptionInfo.Attachment */
+				ptr += sizeof (guint32);
+			}
+
+			if (overrideflags[i] & ARO_SUBTYPE) {
+				/* ExceptionInfo.Subtype */
+				ptr += sizeof (guint32);
+			}
+
+			if (overrideflags[i] & ARO_APPTCOLOR) {
+				/* ExceptionInfo.AppointmentColor */
+				ptr += sizeof (guint32);
+			}
+		}
+
+		/* ExceptionInfo.ReservedBlock1Size */
+		flag32 = *((guint32 *)ptr);
+		ptr += sizeof (guint32);
+		/* the size MUST be zero according to the doc, but... */
+		ptr += flag32;
+
+		/* ExtendedExceptionInfo */
+		for (i = 0; i < count; i++) {
+			/* conditionally parse the ChangeHighlight struct */
+			if (writer_version >= 0x3009) {
+				/* ChangeHighlightSize */
+				flag32 = *((guint32 *)ptr);
+				ptr += sizeof (guint32);
+				/* again, the size MUST be zero according to the doc, but... */
+				ptr += flag32;
+			}
+
+			if (!(overrideflags[i] & (ARO_SUBJECT | ARO_LOCATION)))
+				continue;
+
+			/* ReservedBlockEE1Size */
+			flag32 = *((guint32 *)ptr);
+			ptr += sizeof (guint32);
+			/* again, the size MUST be zero according to the doc, but... */
+			ptr += flag32;
+
+			/* it's supposed to be 0 */
+
+			/* StartTime */
+			ptr += sizeof (guint32);
+
+			/* EndTime */
+			ptr += sizeof (guint32);
+
+			/* OriginalStartDate */
+			ptr += sizeof (guint32);
+
+			if (overrideflags[i] & ARO_SUBJECT) {
+				ECalComponentText text = { 0 };
+				gchar *str;
+
+				/* SubjectLength */
+				flag16 = *(guint16 *)ptr;
+				ptr += sizeof (guint16);
+
+				str = g_convert ((const gchar *) ptr, flag16 * 2, "UTF-8", "UTF-16", NULL, NULL, NULL);
+				text.value = str;
+				e_cal_component_set_summary (detached[i], &text);
+				g_free (str);
+
+				ptr += flag16 * 2;
+			}
+
+			if (overrideflags[i] & ARO_LOCATION) {
+				gchar *str;
+
+				/* LocationLength */
+				flag16 = *(guint16 *)ptr;
+				ptr += sizeof (guint16);
+
+				str = g_convert ((const gchar *) ptr, flag16 * 2, "UTF-8", "UTF-16", NULL, NULL, NULL);
+				e_cal_component_set_location (detached[i], str);
+				g_free (str);
+
+				ptr += flag16 * 2;
+			}
+
+			/* ReservedBlockEE2Size */
+			flag32 = *((guint32 *)ptr);
+			ptr += sizeof (guint32);
+			/* the size MUST be zero according to the doc, but... */
+			ptr += flag32;
+		}
+		for (i = 0; i < count; i++) {
+			*extra_detached = g_slist_append (*extra_detached, detached[i]);
+		}
+		g_free (overrideflags);
+		g_free (detached);
+	}
+
+	/* in case anyone ever needs to traverse further, from this point ptr 
+	   should be pointing at AppointmentRecurrencePattern.ReservedBlock1Size */
 	return TRUE;
 }
 
